/* $Id: scan.c 2379 2022-01-07 22:41:51Z teshields $ */ 
  
/* This file is part of Ox, an Attribute Grammar Compiling System
   that augments the specification functionality of Lex and Yacc, and
   input specification compatible lexer and parser generators.

   Copyright (C) 1992-1993  Kurt M. Bischoff
   Copyright (C) 2016-2022  Thomas E. Shields
   
   This program is free software: you can redistribute it and/or modify
   it under the terms of the GNU General Public License as published by
   the Free Software Foundation, either version 3 of the License, or
   (at your option) any later version.
   
   This program is distributed in the hope that it will be useful,
   but WITHOUT ANY WARRANTY; without even the implied warranty of
   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
   GNU General Public License for more details.
   
   You should have received a copy of the GNU General Public License
   along with this program.  If not, see <http://www.gnu.org/licenses/>.  */

#if defined (__cplusplus)

#include <cstdlib>
#include <cstring>
#include <climits>

#else /* !__cplusplus */

#include <stdio.h> 
#include <stdlib.h> 

#if defined (__sun) && defined (__SVR4)
#include <strings.h> 
#else
#include <string.h> 
#endif
#include <limits.h> 

#endif /* !__cplusplus */

#include "gppl_glob.h"
#include "scan.h"
#include "util.h"

extern int yylineno; 
#define DFAACCEPT 0 

 

void eatComment(void) 
  {char c; 
   int state = 1; 
   while (state != DFAACCEPT)  
     {c = (char)yyyinput();  
      switch (state)   
        {case 1: switch(c) 
                   {case '*' : state = 2; break;  
                    default : ;
                   } 
                 break;  
         case 2: switch(c) 
                   {case '*' : break;  
                    case '/' : state = DFAACCEPT; 
			       break;  
                    default  : state = 1;  
                   }
                 break; 
         default : ;   
        } 
     }  /* while */ 
  } 



#define strBufSize 512

char *copyLexemeOfSTRING(void) 
  {char c; 
   char *cp,*afterBuf,*temp;
   char buf[strBufSize]; 
   int state = 1; 

   cp = buf; 
   *cp++ = '\"';
   afterBuf = buf + strBufSize; 
   while (state != DFAACCEPT)  
     {*cp++ = c = (char)yyyinput();  
      if (cp == afterBuf) internal("string too long\n");  
      switch (state)   
        {case 1: switch(c) 
                   {case '\\' : state = 2; break;  
                    case '\"' : state = DFAACCEPT; break;  
                    default : ;  
                   }
                 break;  
         case 2: state = 1; 
                 break; 
         default : ;   
        } 
     }  /* while */
   if ((temp = (char *)malloc((size_t)(1 + cp - buf))) 
       == 
       ((char *)NULL)
      )
      internal("malloc error in string allocation\n");
   *cp = '\0';
   strcpy(temp,buf); 
   return temp; 
  } 



/* The size of the array of pointers to permanent null-terminated copies 
 * of lexemes of identifiers.  
 */  
static int maxNstrings = 509;  


/* stringTable[i] is either (char *)NULL or a 
 * permanent null-terminated copy of the lexeme of an identifier. 
 */ 
static char **stringTable; 



/*  generates and returns a pointer to a permanent null-terminated copy 
 *  of the first len characters of the string cp.
 */  
static char *allocString(char *cp, int len) 
  {char *ncp, *str_save; 
   if ((ncp = ((char *) malloc((size_t)(len+1)))) 
       == 
       ((char *) NULL) 
      )  
      internal("malloc error in allocString\n"); 
   str_save = ncp; 
   while (len-- != 0) 
         *ncp++ = *cp++; 
   *ncp = '\0';  
   return(str_save); 
  } 



/*  compute a hash value for a string */  
static int stringHash(char *cp, int len) 
  {int hashval; 
   switch (hashval = len) 
     {default: hashval += cp[len - 2] * 512; 
      case 3:  hashval += cp[len - 1] * 64; 
      case 2:  hashval += cp[1] * 8; 
      case 1:  hashval += cp[0]; break;  
      case 0:  internal("can't seek the empty string\n"); 
     }
   return(hashval); 
  } 



/* Search the string table for a string matching the first len characters of
 * cp.  If it is absent and insertion is requested, try to insert it, returning
 * a pointer to a permanent null-terminated copy if successful, null pointer 
 * if unsuccessful (due to a full table).  If absent and insertion not 
 * requested, return null pointer.  If present, return pointer to the copy. 
 */  
char *stringLookup(char *cp, int len, int insertIfAbsent) 
  {int i,itry;  
   char **dum; 
   itry = stringHash(cp,len) % maxNstrings; 
   for (i=0; i<maxNstrings; i++) 
       {dum = stringTable + itry;  
        itry = (itry + 1) % maxNstrings;     
        if (*dum == (char *)NULL) 
           {if (insertIfAbsent != 0) *dum = allocString(cp,len); 
            return(*dum); 
           } 
        if ((int)strlen(*dum) != len) continue; /* don't match proper prefixes */ 
        if (strncmp(*dum,cp,(size_t)len) == 0) return(*dum); 
       } 
   handleOverflow("out of room in string table");  
#if defined(S_SPLINT_S)
   /*@notreached@*/ ;
#else
   /*NOTREACHED*/
#endif
#ifdef __GNUC__
   __builtin_unreachable();
#endif
  } 



void scanSetup(void) 
  {int i;

   if ((stringTable = (char **) malloc((size_t)((size_t)maxNstrings * sizeof(char *))))
       != 
       ((char **) NULL) 
      )   
      for (i=0; i<maxNstrings; i++) stringTable[i] = (char *)NULL; 
      else 
      internal("malloc error in string table allocation\n");  
  }


