/* $Id: bison-directives.cc 2834 2022-12-10 21:33:22Z teshields $ */

/* This file is part of Ox, an Attribute Grammar Compiling System
   that augments the specification functionality of Lex and Yacc, and
   input specification compatible lexer and parser generators.

   Copyright (C) 2021-2022  Thomas E. Shields

   This program is free software: you can redistribute it and/or modify
   it under the terms of the GNU General Public License as published by
   the Free Software Foundation, either version 3 of the License, or
   (at your option) any later version.

   This program is distributed in the hope that it will be useful,
   but WITHOUT ANY WARRANTY; without even the implied warranty of
   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
   GNU General Public License for more details.

   You should have received a copy of the GNU General Public License
   along with this program.  If not, see <http://www.gnu.org/licenses/>.  */

#include "config.h"

#include "location.hh"

#include "bison-directives.hh"
#include "messages.hh"
#include "opts.hh"
#include "scan.hh"
#include "semtypes.hh"



// handle Bison '%language "LANGUAGE"' directive
void
language_directive (const AGC::location& loc, const std::string& literal_string)
{
  yoxControl& ctrl = outMgr::ctrl;

  // extract the language from 'literal_string'
  // 'C' or 'C++' is OK (case insensitive), others not supported

  if (literal_string == "\"C\"" || literal_string == "\"c\"")
    ctrl.targetParserLanguage = languageC;
  else if (literal_string == "\"C++\"" || literal_string == "\"c++\"")
    {
      ctrl.targetParserLanguage = languageCXX;
      ctrl.allLfilesForReentrantParser = true;
    }
  else
    {
      notSupported (loc, "%language " + literal_string);
      ctrl.targetParserLanguage = languageUnknown;
    }
}



// handle Bison '%skeleton "FILE"' directive
void
skeleton_directive (const AGC::location& loc, const std::string& literal_string)
{
  yoxControl& ctrl = outMgr::ctrl;

  // extract the skeleton file name from 'literal_string'
  // '<file>.c' or '<file>.cc' is OK, others not supported

  const int pos = (int)literal_string.find_last_of ('.');
  if (literal_string.substr (pos) == ".c")
    ctrl.targetParserLanguage = languageC;
  else if (literal_string.substr (pos) == ".cc")
    {
      ctrl.targetParserLanguage = languageCXX;
      ctrl.allLfilesForReentrantParser = true;
    }
  else
    {
      notSupported (loc, "%skeleton " + literal_string);
      ctrl.targetParserLanguage = languageUnknown;
    }
}



// handle Bison '%define <API_variable>' directive
void
API_variable_default (const AGC::location& loc, const std::string& API_variable)
{
  yoxControl& ctrl = outMgr::ctrl;

  if (API_variable == "api.pure")
    {
      // "%define api.pure" (default 'true'): generate a reentrant parser
      ctrl.targetParserLanguage = languageC;
      ctrl.allLfilesForReentrantParser = true;
    }
  else if (API_variable == "api.token.constructor")
    {
      // "%define api.token.constructor" (default 'false'):
      // in conjuction with "%define api.value.type variant",
      // don't expect the lexer to return complete symbols,
      // that is, (type, value, and possibly location)
      notSupported (loc, "\"%define api.token.constructor\"");
      ctrl.targetParserLanguage = languageCXX;
      ctrl.allLfilesForReentrantParser = true; // C++ => reentrant parser
    }
  else if (API_variable == "api.value.automove")
    {
      // "%define api.value.automove" (default 'false'):
      // rhs semantic values are not implicitly handled as rvalues
      notSupported (loc, "\"%define api.value.automove\"");
      ctrl.targetParserLanguage = languageCXX;
      ctrl.allLfilesForReentrantParser = true; // C++ => reentrant parser
    }

  // Ox doesn't explicitly care about the following:
  //   api.push-pull (default 'pull')
  //   api.token.raw (default 'false')
  //   lr.default-reduction (default depends on 'canonical-lr' variable)
  //   lr.keep-unreachable-state (default 'false')
  //   lr.type (default 'lalr')
  //   parse.assert (default 'false')
  //   parse.error (default 'simple')
  //   parse.lac (default 'none')
  //   parse.trace (default 'false')
}



// handle Bison '%define <API_variable> ID' directive
void
API_variable_ID (const AGC::location& loc,
                 const std::string& API_variable,
                 const std::string& ID)
{
  yoxControl& ctrl = outMgr::ctrl;

  if (API_variable == "api.pure")
    {
      ctrl.targetParserLanguage = languageC;
      if (ID == "true" || ID == "full")
        // "%define api.pure true", "%define api.pure full":
        // generate a reentrant parser
        ctrl.allLfilesForReentrantParser = true;
      else if (ID == "false")
        // "%define api.pure false": generate a non-reentrant parser
        ctrl.allLfilesForReentrantParser = false;
    }
  else if (API_variable == "api.token.constructor")
    {
      // "%define api.token.constructor true",
      // "%define api.token.constructor false":
      // if 'true', in conjuction with "%define api.value.type variant",
      // expect the lexer to return complete symbols,
      // that is, (type, value, and possibly location)
      notSupported (loc, "\"%define api.token.constructor\"");
      ctrl.targetParserLanguage = languageCXX;
      ctrl.allLfilesForReentrantParser = true; // C++ => reentrant parser
    }
  else if (API_variable == "api.value.automove")
    {
      // "%define api.value.automove true",
      // "%define api.value.automove false":
      // if 'true', rhs semantic values are implicitly handled as rvalues
      notSupported (loc, "\"%define api.value.automove\"");
      ctrl.targetParserLanguage = languageCXX;
      ctrl.allLfilesForReentrantParser = true; // C++ => reentrant parser
    }
  else if (API_variable == "api.value.type")
    {
      if (ID == "union")
        {
          // "%define api.value.type union":
          // generate a '%union' declaration from the '%token' declarations
          if (ctrl.targetParserGen == bisonParserGen &&
              (bisonApiValueTypeUnsupported > 0 ||
               bisonApiValueTypeUnionDefined))
            multipleYYSTYPEdefinitions (loc, "api.value.type");

          if (yaccUnionDefined > 0)
            conflictingYYSTYPEdefinitions (loc,
                                           "%define api.value.type union",
                                           "%union");

          bisonApiValueTypeUnionDefined = true;
        }
      else if (ID == "union-directive")
        {
          // "%define api.value.type union-directive":
          // use the Yacc '%union' declaration
          if (ctrl.targetParserGen == bisonParserGen &&
              (bisonApiValueTypeUnsupported > 0 ||
               bisonApiValueTypeUnionDefined))
            multipleYYSTYPEdefinitions (loc, "api.value.type");
        }
      else if (ID == "variant")
        {
          // "%define api.value.type variant":
          // generate a virtual '%union' declaration from
          // the '%token' declarations

// suppress multiple errors for unsupported "%define" directives
//        if (ctrl.targetParserGen == bisonParserGen &&
//            (bisonApiValueTypeUnsupported > 0 ||
//             bisonApiValueTypeUnionDefined))
//          multipleYYSTYPEdefinitions (loc, "api.value.type");

          notSupported (loc, "\"%define api.value.type variant\"");
          bisonApiValueTypeUnsupported++;

// suppress multiple errors for unsupported '%define' directives
//        if (yaccUnionDefined > 0)
//          conflictingYYSTYPEdefinitions (loc,
//                                         "%define api.value.type variant",
//                                         "%union");

          ctrl.targetParserLanguage = languageCXX;
          ctrl.allLfilesForReentrantParser = true; // C++ => reentrant parser
        }
    }

  // Ox doesn't explicitly care about the following:
  //   api.push-pull pull|push|both => C
  //   api.value.union.name ID => C
  //   api.location.file none => C++
  //   api.token.raw true|false
  //   lr.type canonical-lr
  //   lr.default-reduction most|consistent|accepting
  //   lr.keep-unreachable-state true|false
  //   lr.type lalr|ielr
  //   parse.assert true|false
  //   parse.error simple|detailed|verbose|custom
  //   parse.lac none|full
  //   parse.trace true|false
}



// handle Bison '%define <API_variable> LITERAL' directive
void
API_variable_LITERAL (const AGC::location& loc __attribute__ ((unused)),
                      const std::string& API_variable,
                      const std::string& LITERAL)
{
  yoxControl& ctrl = outMgr::ctrl;

  if (API_variable == "api.prefix")
    {
      ctrl.yaccSymbolPrefix = LITERAL;
    }
  else if (API_variable == "name-prefix" || API_variable == "name_prefix")
    {
      ctrl.yaccSymbolPrefix = LITERAL;
      ctrl.user_namespace = LITERAL;
    }

  // Ox doesn't explicitly care about the following:
  //   api.header.include { "FILE" } => C
  //   api.header.include { <PATH_NAME> } => C
  //   api.location.file "FILE" => C++
  //   api.location.include { "FILE" } => C++
  //   api.location.include { <PATH_NAME> } => C++
}



// handle 'bison' directives:
//      '%define <API_variable> { ID }'
//      '%define <API_variable> { NAMESPACE_ID }'
//      '%define <API_variable> { <type> }'
int
API_variable_EXPR (const AGC::location& loc,
                   const std::string& API_variable,
                   const typeSeq EXPR)
{
  yoxControl& ctrl = outMgr::ctrl;

  if (EXPR.empty ()) // 'api.value.type { }' is not currently supported by bison
    return (0);

  int num_values =
      (int)EXPR.size (); // the number of values provided should be exactly 1

  if (API_variable == "api.parser.class" || API_variable == "parser_class_name")
    {
      if (num_values == 1 && // verify 'EXPR' is an 'ID'
          EXPR.front ().kind == is_ID)
        // "%define api.parser.class {ID}", "%define parser_class_name {ID}":
        // specify the C++ parser class name
        ctrl.user_parser = EXPR.front ().item; // generated parser class name

      ctrl.targetParserLanguage = languageCXX;
      ctrl.allLfilesForReentrantParser = true; // C++ => reentrant parser
    }
  else if (API_variable == "api.prefix")
    {
      if (num_values == 1 && // verify 'EXPR' is an 'ID'
          EXPR.front ().kind == is_ID)
        {
          // "%define api.prefix {ID}":
          // specify the replacement for "yy" & "YY"
          ctrl.yaccSymbolPrefix = EXPR.front ().item; // prefix replacing "yy"
        }
    }
  else if (API_variable == "api.token.prefix")
    {
      if (num_values == 1 && // verify 'EXPR' is an 'ID'
          EXPR.front ().kind == is_ID)
        {
          // "%define api.token.prefix {ID}":
          // specify the prefix to add to the token names
          ctrl.user_token_prefix =
              EXPR.front ().item; // prefix to lexer token names
        }
    }
  else if (API_variable == "api.namespace" || API_variable == "namespace")
    {
      if (num_values == 1 && // verify 'EXPR' is a 'NAMESPACE_ID'
          EXPR.front ().kind == is_NAMESPACE_ID)
        // "%define api.namespace {NAMESPACE_ID}",
        // "%define namespace {NAMESPACE_ID}":
        // specify the namespace for the C++ parser class
        ctrl.user_namespace =
            EXPR.front ().item; // generated parser class namespace

      ctrl.targetParserLanguage = languageCXX;
      ctrl.allLfilesForReentrantParser = true; // C++ => reentrant parser
    }
  else if (API_variable == "api.value.type")
    {
      // "%define api.value.type {<type>}":
      // ignore any Yacc '%union' declaration
      // use <type> as the semantic stack element type

// suppress multiple errors for unsupported "%define" directives
//    if (ctrl.targetParserGen == bisonParserGen &&
//        (bisonApiValueTypeUnsupported > 0 ||
//         bisonApiValueTypeUnionDefined))
//      multipleYYSTYPEdefinitions (loc, "api.value.type");

      notSupported (loc, "\"%define api.value.type {TYPE}\"");
      bisonApiValueTypeUnsupported++;

// suppress multiple errors for unsupported "%define" directives
//    if (yaccUnionDefined > 0)
//      conflictingYYSTYPEdefinitions (loc,
//                                     "%define api.value.type {TYPE}",
//                                     "%union");
    }

  return (num_values);

  // Ox doesn't explicitly care about the following:
  //   api.symbol.prefix { ID }
  //   api.filename.type { <type> } => C++
  //   api.location.type { <type> }
  //   api.value.type { } => not currently supported by bison
}
