/* $Id: agc-Yfile-lexer.hh 2834 2022-12-10 21:33:22Z teshields $ */

/* This file is part of Ox, an Attribute Grammar Compiling System
   that augments the specification functionality of Lex and Yacc, and
   input specification compatible lexer and parser generators.

   Copyright (C) 2022  Thomas E. Shields

   This program is free software: you can redistribute it and/or modify
   it under the terms of the GNU General Public License as published by
   the Free Software Foundation, either version 3 of the License, or
   (at your option) any later version.

   This program is distributed in the hope that it will be useful,
   but WITHOUT ANY WARRANTY; without even the implied warranty of
   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
   GNU General Public License for more details.

   You should have received a copy of the GNU General Public License
   along with this program.  If not, see <http://www.gnu.org/licenses/>.  */

#ifndef AGC_YFILE_LEXER_HH_INCLUDED
#define AGC_YFILE_LEXER_HH_INCLUDED

#if !defined(yyFlexLexer)
#define yyFlexLexer Yfile_FlexLexer
#include <FlexLexer.h>

// 'flex' secrets:
#define BEGIN (yy_start) = 1 + 2 *
#define YY_START (((yy_start) - 1) / 2)
#endif

#include <fstream>
#include <stack>

#include "agc-Yfile.hh"
#include "mac.hh"
#include "outMgr.hh"
#include "scan.hh"

namespace AGC
{
namespace YFILE
{

class Lexer : public Yfile_FlexLexer
{
public:

  Lexer (Yfile_specIO&, std::ifstream&);

  inline int
  yylex (parser::value_type* yylval, location* yylloc)
  {
    return yylex (*yylval, *yylloc);
  }

  int yylex (parser::value_type& yylval, location& yylloc);

  inline int
  yylex (void) // should never be called
  {
    return parser::symbol_kind::S_YYerror;
  }

  inline int&
  lineno (void)
  {
    return yylineno;
  }

  inline int&
  columno (void)
  {
    return yycolumn;
  }

protected:

  inline int
  start (void) const
  {
    return YY_START;
  }

  inline void
  start (int new_state)
  {
    BEGIN (new_state);
  }

  inline size_t
  size (void) const
  {
    return (size_t)yyleng;
  }

  inline int
  chr (void) const
  {
    return (int)*yytext;
  }

  inline const char*
  text (void) const
  {
    return yytext;
  }

  inline const std::string
  str (void) const
  {
    return std::string (yytext);
  }

  inline void
  push_state (const position& pos, int new_state)
  {
    yy_push_state (new_state);
    startPos.push (pos); // starting position if EOF in 'new_state'
  }

  inline void
  push_state (const location& loc, int new_state)
  {
    yy_push_state (new_state);
    startPos.push (loc.begin); // starting position if EOF in 'new_state'
  }

  inline void
  pop_state (void)
  {
    yy_pop_state ();
    startPos.pop ();
  }

#if defined(_WIN32)
  int LexerInput (char*, int); // overload to treat '\r' or "\r\n" as '\n'
#endif

  Yfile_specIO& theFiles;

  EOLctx EOLstate;

private:

  // 'actionCLOSER' holds the closing '}', any surrounding comments and
  // whitespace (to // include '\n') and trailing rule directive context, of a
  // Yacc rule action, until it is known if this is the last action.

  // After the closing curly brace is detected in state YACC_ACT, it has to
  // be determined whether the action just scanned is the last thing other than
  // an attribute reference section, vertical bar, or semicolon, in the current
  // rule body (RHS).
 
  // If it is the last thing, then code that generates a parse tree node (and
  // possibly evaluates attributes computable during the bottom-up parse) upon
  // reduction by the production is inserted before the closing brace of the
  // action is output.  If the last thing in a rule body isn't an action, then
  // Ox generates it's own final rule action.

  // Note: AT&T Yacc doesn't permit two adjacent actions in a given rule body.
  inline void
  notLastSemAction (void)
  {
    yoxOutG.append (actionCLOSER);
    actionCLOSER.clear ();
  }

  // 'possibleLHS' may hold a rule LHS (nonterminal symbol, inherited attributes
  // if BtYacc or BYacc is the target parser generator, ':', possibly plus '|'
  // if the initial RHS is empty or ';' if the only RHS is empty), including any
  // surrounding comments and whitespace; any immediately following symbol
  // indicates that it doesn't hold a rule LHS.
 
  // If 'possibleLHS' holds the rule LHS, the contents are output after the
  // Ox-generated start rule has been generated; otherwise, the contents are
  // output immediately.
  inline void
  notLHS (void)
  {
    yoxOutG.append (possibleLHS);
    possibleLHS.clear ();
  }

  int yycolumn;

  int nLbraces; // '{' ... '}' nesting
  int nLTbraces; // '<' ... '>' nesting
  int nLbrackets; // '[' ... ']' nesting
  int nLparens; // '(' ... ')' nesting
  int nLparensM; // macro invocation '(' ... ')' nesting
  int nATLbraces; // check for unexpected nested "@{"

  outMgr* whereTo;
  void
  yox_push_state (int, outMgr&);

  std::string WSbuffer;

  void
  bufferWS (void)
  {
    WSbuffer = YYText ();
  }

  void
  flushWS (outMgr& whTo)
  {
    if (!WSbuffer.empty ())
      {
        whTo.lex ().put (WSbuffer);
        WSbuffer.clear ();
      }
  }

  std::stack<position> startPos;

  bool stringHasBareNewline;

  macroArg macroARG;

  yoxControl& ctrl;

  bool pass2parser = false; // true => pass '%define' directive
                            // argument(s) to the parser

  std::ostringstream formatted_msg;
};

}
}



// begin hack: required by broken, IMHO, 'flex' C++ lexer implememtation

// There is no way to override the signature of 'yylex()' in the base
// classes 'FlexLexer' and 'yyFlexLexer'.

// A Bison C++ parser is reentrant and expects the following signature:
//     int yylex (YYSTYPE* yylval, YYLTYPE* yylloc)

// The 'YY_DECL' macro can define the signature of the 'yylex ()' member
// function in an application lexer derived from 'yyFlexLexer', leaving
// the 'yyFlexLexer::yylex (void)' member function hanging out in the cold;
// here is it:

inline int // no function actually generated
yyFlexLexer::yylex (void)
{
  return 0; // never called
}

// end hack

#endif // !AGC_YFILE_LEXER_HH_INCLUDED
