/* $Id: main.cc 2834 2022-12-10 21:33:22Z teshields $ */

/* This file is part of Ox, an Attribute Grammar Compiling System
   that augments the specification functionality of Lex and Yacc, and
   input specification compatible lexer and parser generators.

   Copyright (C) 2021-2022  Thomas E. Shields

   This program is free software: you can redistribute it and/or modify
   it under the terms of the GNU General Public License as published by
   the Free Software Foundation, either version 3 of the License, or
   (at your option) any later version.

   This program is distributed in the hope that it will be useful,
   but WITHOUT ANY WARRANTY; without even the implied warranty of
   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
   GNU General Public License for more details.

   You should have received a copy of the GNU General Public License
   along with this program.  If not, see <http://www.gnu.org/licenses/>.  */

#include "config.h"

#include <ios>
#include <sstream>

#if defined(WITH_REFLEX)
#include <reflex/input.h>
#endif

#if HAVE_LTDL_H
#include <ltdl.h>
#endif

#include "fileMgr.hh"
#include "joinTabs.hh"
#include "lextypes.hh"
#include "messages.hh"
#include "opts.hh"
#include "outMgr.hh"
#include "scan.hh"
#include "strip.hh"
#include "trav.hh"
#include "treeviz.hh"


#if defined(USE_LT_DLOPEN)
#if defined(WITH_REFLEX)
typedef void (*dl_parser) (const reflex::Input&, Yfile_specIO&, int);
#else // defined(WITH_FLEX)
typedef void (*dl_parser) (std::ifstream&, Yfile_specIO&, int);
#endif

static dl_parser parse = nullptr;

static lt_dlhandle parser_handle;



static void
load_agc_Yfile_parser ()
{
  const yoxControl& ctrl = outMgr::ctrl;

  std::string parser_library ("libox_");
  std::string yyparse_name ("agc_");

  if (lt_dlinit () != 0) // `lt_dlinit ()' failed
    {
      AGC::location const loc; // dummy location
      yox_fatal (loc, std::string (lt_dlerror ()) + ".\n");
    }

  // compute the library name
  switch (ctrl.targetParserGen)
    {
    case bisonParserGen:
    case bisonParserGenOld:
      parser_library.append ("bison.la");
      yyparse_name.append ("bison_parse");
      break;
    case btyaccParserGen:
      parser_library.append ("btyacc.la");
      yyparse_name.append ("btyacc_parse");
      break;
    case byaccParserGen:
      parser_library.append ("byacc.la");
      yyparse_name.append ("byacc_parse");
      break;
    case mstaParserGen:
      parser_library.append ("msta.la");
      yyparse_name.append ("msta_parse");
      break;
    case yaccParserGen:
      parser_library.append ("yacc.la");
      yyparse_name.append ("yacc_parse");
      break;
    }

  parser_handle = lt_dlopen (parser_library.c_str ());
  if (parser_handle == nullptr) // didn't find the parser library
    {
      AGC::location const loc; // dummy location
      yox_fatal (loc,
                 std::string (lt_dlerror ()) + ": " + parser_library + ".\n");
    }

  parse = reinterpret_cast<dl_parser> (
      lt_dlsym (parser_handle, yyparse_name.c_str ()));
  if (parse == nullptr) // didn't find the `yyparse' entry point
    {
      AGC::location const loc; // dummy location
      yox_fatal (loc, std::string (lt_dlerror ()) + ": yyparse ().\n");
    }
}



static void
unload_agc_Yfile_parser (void)
{
  if (lt_dlclose (parser_handle) != 0) // `lt_dlclose ()' failed
    {
      AGC::location const loc; // dummy location
      yox_fatal (loc, std::string (lt_dlerror ()) + ".\n");
    }

  if (lt_dlexit () != 0) // `lt_dlexit ()' failed
    {
      AGC::location const loc; // dummy location
      yox_fatal (loc, std::string (lt_dlerror ()) + ".\n");
    }
}
#endif


#if BISON_TARGET_ONLY || ALL_TARGETS_STATIC || defined(USE_LT_DLOPEN)
#include "agc-bison.hh"
#endif
#if BTYACC_TARGET_ONLY || ALL_TARGETS_STATIC || defined(USE_LT_DLOPEN)
#include "agc-btyacc.hh"
#endif
#if BYACC_TARGET_ONLY || ALL_TARGETS_STATIC || defined(USE_LT_DLOPEN)
#include "agc-byacc.hh"
#endif
#if MSTA_TARGET_ONLY || ALL_TARGETS_STATIC || defined(USE_LT_DLOPEN)
#include "agc-msta.hh"
#endif
#if YACC_TARGET_ONLY || ALL_TARGETS_STATIC || defined(USE_LT_DLOPEN)
#include "agc-yacc.hh"
#endif


static void
#if defined(WITH_REFLEX)
agc_Yfile_parse (const reflex::Input& yoxIn, Yfile_specIO& Yfile, int debug)
#else // defined(WITH_FLEX)
agc_Yfile_parse (std::ifstream& yoxIn, Yfile_specIO& Yfile, int debug)
#endif
{
#if defined(USE_LT_DLOPEN)
  parse (yoxIn, Yfile, debug);
#else
#if ALL_TARGETS_STATIC
  switch (outMgr::ctrl.targetParserGen)
    {
    case bisonParserGen:
    case bisonParserGenOld:
#endif
#if BISON_TARGET_ONLY | ALL_TARGETS_STATIC
      agc_bison_parse (yoxIn, Yfile, debug);
#endif
#if ALL_TARGETS_STATIC
      return;
    case btyaccParserGen:
#endif
#if BTYACC_TARGET_ONLY || ALL_TARGETS_STATIC
      agc_btyacc_parse (yoxIn, Yfile, debug);
#endif
#if ALL_TARGETS_STATIC
      return;
    case byaccParserGen:
#endif
#if BYACC_TARGET_ONLY || ALL_TARGETS_STATIC
      agc_byacc_parse (yoxIn, Yfile, debug);
#endif
#if ALL_TARGETS_STATIC
      return;
    case mstaParserGen:
#endif
#if MSTA_TARGET_ONLY || ALL_TARGETS_STATIC
      agc_msta_parse (yoxIn, Yfile, debug);
#endif
#if ALL_TARGETS_STATIC
      return;
    case yaccParserGen:
#endif
#if YACC_TARGET_ONLY || ALL_TARGETS_STATIC
      agc_yacc_parse (yoxIn, Yfile, debug);
#endif
#if ALL_TARGETS_STATIC
      return;
    }

#if defined(__GNUC__)
  __builtin_unreachable ();
#endif
#endif
#endif // defined(USE_LT_DLOPEN)
}


#include "agc-Lfile.hh"

static void
closeAndEraseIOfiles (void)
{
  closeIOfiles ();
  eraseIOfiles ();
}

#if defined(USE_BDWGC)
#include <gc/gc.h>
#endif

int
main (int argc, char* argv[])
{
  char* ENV_YYDEBUG = getenv ("YYDEBUG");
  // YYDEBUG=0  => no tracing
  //         1  => parser trace
  //         2  => lexer trace
  //         3  => lexer & parser trace

  int debug = 0;

  if (ENV_YYDEBUG != nullptr)
    {
      try
        {
          debug = std::stoi (ENV_YYDEBUG);
        }
      catch (...)
        {
          // debug = 0;
        }
      if (debug < 0 || debug > 3)
        debug = 1; // => parser trace
    }

  setupNIOsym (); // initialize the Node-Image Object pseudo attribute name

  setOptions (argc, argv);

  yoxControl& ctrl = outMgr::ctrl;

  openIOfiles ();

  // call at exit
  if (ctrl.saveBadOutputs)
    (void)atexit (&closeIOfiles);
  else
    (void)atexit (&closeAndEraseIOfiles);

  if (ctrl.stripOxConstructs)
    {
      stripInputFiles (debug);
      return (EXIT_SUCCESS);
    }

  openTempFiles ();
  (void)atexit (&rmTempFiles); // call at exit

  prepareDataStructures2 ();

#if defined(USE_LT_DLOPEN)
  load_agc_Yfile_parser ();
#endif

  // open the input Y-file
#if defined(WITH_REFLEX) && defined(_WIN32)
  FILE* yoxInGF = fopen (Yfile.inName.c_str (), "rb");
  if (yoxInGF == nullptr)
#else // defined(WITH_FLEX) || defined(WITH_REFLEX) && !defined(_WIN32)
  std::ifstream yoxInG (Yfile.inName);
  if (!yoxInG.is_open ())
#endif
    {
      const AGC::location loc (&Yfile.inName);
      yox_fatal (loc, "opening " + Yfile.inName + " for input failed\n");
    }

#if defined(WITH_REFLEX) && defined(_WIN32)
  reflex::BufferedInput::dos_streambuf buf (yoxInGF); // "\r\n" -> '\n'
  std::istream yoxInG (&buf);
  if (!yoxInG.good ())
    {
      const AGC::location loc (&Yfile.inName);
      yox_fatal (loc, "opening " + Yfile.inName + " for CRLF->LF failed\n");
    }
#endif

  // open the output Y-file
  yoxOutG.open (Yfile.outName);

  agc_Yfile_parse (yoxInG, Yfile, debug);

  // close the input Y-file
#if defined(WITH_REFLEX) && defined(_WIN32)
  fclose (yoxInGF);
#else // defined(WITH_FLEX) || defined(WITH_REFLEX) && !defined(_WIN32)
  yoxInG.close ();
#endif

#if defined(USE_LT_DLOPEN)
  unload_agc_Yfile_parser ();
#endif

  for (Lfile_specIO& currentLfile : Lfile)
    {
      switch (currentLfile.scanner)
        {
        case LEX_SPEC_LEXER:
          // set the language and reentrant parser defaults for this lexer
          ctrl.targetLexerLanguageForThisLfile = ctrl.targetLexerLanguage;
          ctrl.thisLfileForReentrantParser = ctrl.allLfilesForReentrantParser;
          break;
        case GEN_SPEC_LEXER:
          break;
        default:
          yox_internal_fail (__FILE__, __LINE__,
                             "unexpected Lexer " +
                                 std::to_string (currentLfile.scanner) + "\n");
        }

        // open the (next) input L-file
#if defined(WITH_REFLEX) && defined(_WIN32)
      FILE* yoxInLF = fopen (currentLfile.inName.c_str (), "rb");
      if (yoxInLF == nullptr)
#else // defined(WITH_FLEX) || defined(WITH_REFLEX) && !defined(_WIN32)
      std::ifstream yoxInL (currentLfile.inName);
      if (!yoxInL.is_open ())
#endif
        {
          const AGC::location loc (&currentLfile.inName);
          yox_fatal (loc,
                     "opening " + currentLfile.inName + " for input failed\n");
        }

#if defined(WITH_REFLEX) && defined(_WIN32)
      reflex::BufferedInput::dos_streambuf buf (yoxInLF); // "\r\n" -> '\n'
      std::istream yoxInL (&buf);
      if (!yoxInL.good ())
        {
          const AGC::location loc (&currentLfile.inName);
          yox_fatal (loc,
                     "opening " + currentLfile.inName +
                         " for CRLF->LF failed\n");
        }
#endif

      if (currentLfile.scanner == GEN_SPEC_LEXER)
        yoxOutL = fileMgr (false); // -> Gen lexer output

      // open the (next) output L-file
      yoxOutL.open (currentLfile.outName);

      yoxOutL.putLexerHeader (currentLfile.scanner);

      agc_lexer_parse (yoxInL, currentLfile, debug);

      // close the current output L-file
      yoxOutL.close ();

      // close the current input L-file
#if defined(WITH_REFLEX) && defined(_WIN32)
      fclose (yoxInLF);
#else // defined(WITH_FLEX) || defined(WITH_REFLEX) && !defined(_WIN32)
      yoxInL.close ();
#endif
    }

  organizeTables ();

  if (ctrl.removeOutput)
    return (EXIT_FAILURE);
  else
    return (EXIT_SUCCESS);
}
