/* $Id: opts.cc 2866 2022-12-19 22:24:52Z teshields $ */

/* This file is part of Ox, an Attribute Grammar Compiling System
   that augments the specification functionality of Lex and Yacc, and
   input specification compatible lexer and parser generators.

   Copyright (C) 1992-1994  Kurt M. Bischoff
   Copyright (C) 2004-2022  Thomas E. Shields

   This program is free software: you can redistribute it and/or modify
   it under the terms of the GNU General Public License as published by
   the Free Software Foundation, either version 3 of the License, or
   (at your option) any later version.

   This program is distributed in the hope that it will be useful,
   but WITHOUT ANY WARRANTY; without even the implied warranty of
   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
   GNU General Public License for more details.

   You should have received a copy of the GNU General Public License
   along with this program.  If not, see <http://www.gnu.org/licenses/>.  */

#include "config.h"

#include <fstream>

#include <climits>
#include <cstring>

#include <getopt.h>

#include "location.hh"

#include "ctc.hh"
#include "fileMgr.hh"
#include "messages.hh"
#include "opts.hh"
#include "outMgr.hh"
#include "patchlvl.hh"
#include "scan.hh"
#include "treeviz.hh"


static std::string yoxInFileStrG;

static const std::string yoxOutFileDefaultBaseName ("oxout");



static void
printPatchlevel (std::ostream& out)
{
  out << PATCHLEVEL << "\n";
}



static void
printVDATE (std::ostream& out)
{
  out << VDATE << "\n";
}



static const char* const usageText =
    "\n"
    "Usage: ox {option}* [--] `Y-file\' {[-G] `L-file\'}+\n"
    "  Y-file                   attribute augmented parser specification\n"
    "  L-file                   attribute augmented lexer specification\n"
    "option:\n"
    "  --help                   show this usage summary and exit\n"
    "  -b, --prefix=basename    use \"basename\" instead of \"oxout\"\n"
    "  -c                       suppress cycle reports\n"
    "  -d                       suppress pruning\n"
    "  -G L-file                process `L-file\' as a non-Lex lexer file\n"
    "  -h                       build header file\n"
    "  -j                       save bad output files\n"
#if WITH_LINE_DIRECTIVES
    "  -L, -l, --noline         suppress `#line\' directives\n"
#else
    "  -L, -l, --noline         default: suppress `#line\' directives\n"
#endif
    "  -n                       neutralize semantic actions\n"
    "  -p, --name-prefix=prefix use \"prefix\" instead of \"yyy\" for external "
        "symbols\n"
    "  -S, --strip              strip Ox constructs\n"
    "  -t                       build internal visual representation of parse "
        "tree\n"
    "  -u                       show attributed parse tree size statistics\n"
    "  -V, --version            show version information and exit\n"
    "  --header-prefix=basename use \"basename\" instead of \"oxout\" for "
        "header\n"
    "  --lex-prefix=basename    use \"basename\" instead of \"oxout\" for "
        "lexer\n"
    "  --parse-prefix=basename  use \"basename\" instead of \"oxout\" for "
        "parser\n"
    "  --reentrant-parser, --pure-parser\n"
    "                           compile `Y-file\' assuming it is a reentrant\n"
    "                           (pure) parser; compile each `L-file\' "
        "accordingly\n"
    "  --language=[C|C++]       compile `Y-file\' and each Lex `L-file\' with "
        "C or C++\n"
    "                           as the target language (default: "
        "--language=C)\n"
#if BISON_TARGET_ONLY || ALL_TARGETS_STATIC || defined(USE_LT_DLOPEN)
    "  --bison[=C|C++]          compile `Y-file\' for `bison\' in C (default) "
        "or C++;\n"
    "                           if C++ then \"bison --language=c++\" (or "
        "equivalent)\n"
    "                           must be used (implies \"--reentrant-parser\")\n"
    "  --bison-2.3[=C|C++]      same as \"--bison\", for `bison\' v2.3 "
        "functionality\n"
#endif
#if BTYACC_TARGET_ONLY || ALL_TARGETS_STATIC || defined(USE_LT_DLOPEN)
    "  --btyacc[=C|C++]         compile `Y-file\' for `btyacc\' in C (default) "
        "or C++\n"
#endif
#if BYACC_TARGET_ONLY || ALL_TARGETS_STATIC || defined(USE_LT_DLOPEN)
    "  --byacc[=C|C++]          compile `Y-file\' for `byacc\' in C (default) "
        "or C++\n"
#endif
#if MSTA_TARGET_ONLY || ALL_TARGETS_STATIC || defined(USE_LT_DLOPEN)
    "  --msta[=C|C++]           compile `Y-file\' for `msta\' in C (default) "
        "or C++;\n"
    "                           if C++ then \"msta -c++\" must be used\n"
#endif
#if YACC_TARGET_ONLY || ALL_TARGETS_STATIC || defined(USE_LT_DLOPEN)
    "  --yacc[=C|C++]           compile `Y-file\' for `yacc\' functionality in "
        "C (this\n"
    "                           is the default for a `Y-file') or C++\n"
#endif
    "  --flex[=C|C++]           compile each Lex `L-file\' for `flex\' in C "
        "(default)\n"
    "                           or C++; if C++ then \"flex --c++\" (or "
        "equivalent)\n"
    "                           must be used\n"
    "  --lex[=C|C++]            compile each Lex `L-file\' for `lex\' "
        "functionality\n"
    "                           in C (this is the default for a Lex `L-file') "
        "or C++\n"
    "  --reflex[=C++]           compile each Lex `L-file\' for `reflex\' in "
        "C++\n";



static void
printUsageText (std::ostream& out)
{
  out << usageText;
}



static void
addFilenameExt (std::string& outName, std::string& inName, char d)
{
  std::string ext (1, '.');
  int last_dot = (int)inName.find_last_of ('.');
  std::string inNameExt;

  if (last_dot != ((long)std::string::npos) &&
      (((long)inName.length ()) > (last_dot + 1)))
    inNameExt = inName.substr (last_dot + 1);

  if (inNameExt.length () <= 1)
    {
      // no extension, or a 1-char extension; use the default
      ext.append (1, d);
    }
  else
    {
      // multi-char extension, assume C++ filename conventions:
      //  - if exactly 2 identical characters, replace both by the default
      //  - othersize, convert the input extension to lower case, and
      //    replace the first character with the default

      if ((inNameExt.length () == 2) && (inNameExt[0] == inNameExt[1]))
        ext.append (2, d);
      else
        {
          ext.append (1, d); // new extension starts with the default
          for (int i = 1; i < (long)inNameExt.length (); i++)
            ext.append (1, (char)tolower (inNameExt[i]));
        }
    }

  outName.append (ext);
}



enum long_opts
{
  HEADER_PREFIX_OPT = (UCHAR_MAX + 1),
  HELP_OPT,
  LEX_PREFIX_OPT,
  PARSE_PREFIX_OPT,
  PURE_PARSER_OPT,
  LANGUAGE_OPT,
  FLEX_OPT,
  REFLEX_OPT,
  LEX_OPT,
  BISON_OPT,
  BISON_OLD_OPT,
  BTYACC_OPT,
  BYACC_OPT,
  MSTA_OPT,
  YACC_OPT,
};

// options descriptor
static struct option opts[] = {
  { "help",             no_argument,       nullptr, HELP_OPT },
  { "prefix",           required_argument, nullptr, 'b' },
  { "noline",           no_argument,       nullptr, 'L' },
  { "name-prefix",      required_argument, nullptr, 'p' },
  { "strip",            no_argument,       nullptr, 'S' },
  { "version",          no_argument,       nullptr, 'V' },
  { "header-prefix",    required_argument, nullptr, HEADER_PREFIX_OPT },
  { "lex-prefix",       required_argument, nullptr, LEX_PREFIX_OPT },
  { "parse-prefix",     required_argument, nullptr, PARSE_PREFIX_OPT },
  { "reentrant-parser", no_argument,       nullptr, PURE_PARSER_OPT },
  { "pure-parser",      no_argument,       nullptr, PURE_PARSER_OPT },
  { "language",         required_argument, nullptr, LANGUAGE_OPT },
  { "bison",            optional_argument, nullptr, BISON_OPT },
  { "bison-2.3",        optional_argument, nullptr, BISON_OLD_OPT },
  { "btyacc",           optional_argument, nullptr, BTYACC_OPT },
  { "byacc",            optional_argument, nullptr, BYACC_OPT },
  { "msta",             optional_argument, nullptr, MSTA_OPT },
  { "yacc",             optional_argument, nullptr, YACC_OPT },
  { "flex",             optional_argument, nullptr, FLEX_OPT },
  { "reflex",           optional_argument, nullptr, REFLEX_OPT },
  { "lex",              optional_argument, nullptr, LEX_OPT },
  { nullptr, 0, nullptr, 0 }
};


static std::string yoxOutFileStrHbaseName;



yoxControl outMgr::ctrl;



static const std::string
make_upper (const std::string& name)
{
  std::string upper = name;
  for (char& c : upper)
    if (islower (c))
      c = (char)toupper (c);
    else if (!isalpha (c))
      c = '_';
  return upper;
}



const std::string
yaccPrefix (const std::string base, bool uc)
{
  if (uc)
    return make_upper (outMgr::ctrl.yaccSymbolPrefix) + base;
  else
    return outMgr::ctrl.yaccSymbolPrefix + base;
}



const std::string
yoxPrefix (const std::string base)
{
  return outMgr::ctrl.yoxSymbolPrefix + base;
}



const std::string
addTokenPrefix (const std::string base)
{
  return outMgr::ctrl.user_token_prefix + base;
}



const std::string
stripTokenPrefix (const std::string token)
{
  yoxControl& ctrl = outMgr::ctrl;

  if (ctrl.user_token_prefix.length () > 0 &&
      ctrl.user_token_prefix ==
          token.substr (0, ctrl.user_token_prefix.length ()))
    return token.substr (ctrl.user_token_prefix.length ());
  else
    return token;
}



// check the Flex & RE/flex '%option' directive for those of interest to Ox
void
checkLexGenOptionDirective (const AGC::location& loc, std::string options)
{
  yoxControl& ctrl = outMgr::ctrl;

  ctrl.targetLexerGen = flexLexerGen;

  do
    {
      size_t next_pos = options.find_first_of (" \t\f");

      const std::string token = options.substr (0, next_pos);

      if (token.empty ())
        ; // no-op
      else if (token == "bison-bridge")
        ctrl.thisLfileForReentrantParser = true;
      else if (token == "nobison-bridge")
        ctrl.thisLfileForReentrantParser = false;
      else if (token == "bison-locations")
        ctrl.thisLfileForReentrantParser = true;
      else if (token == "nobison-locations")
        ctrl.thisLfileForReentrantParser = false;
      else if (token == "reentrant")
        ctrl.thisLfileForReentrantParser = true;
      else if (token == "noreentrant")
        ctrl.thisLfileForReentrantParser = false;
      else if (token == "bison-complete")
        {
          // RE/flex unique option

          ctrl.targetLexerGenForThisLfile = reflexLexerGen;
          ctrl.thisLfileForReentrantParser = true;
          ctrl.targetLexerLanguageForThisLfile = languageCXX;

          // requires 'bison' directives "%define api.value.type variant"
          // and "%define api.token.constructor", which are not supported
          notSupported (loc, "\"%option bison-complete\"");
        }
      else if (token == "c++")
        ctrl.targetLexerLanguageForThisLfile = languageCXX;

      if (next_pos == std::string::npos)
        break;

      options = options.substr (next_pos + 1);
    }
  while (!options.empty ());
}



// set '*ForThisLexer' inferences for RE/flex unique directives
void
reflexDirective (void)
{
  yoxControl& ctrl = outMgr::ctrl;

  ctrl.targetLexerGenForThisLfile = reflexLexerGen;
  ctrl.thisLfileForReentrantParser = true;
  ctrl.targetLexerLanguageForThisLfile = languageCXX;
}



void
setOptions (int argc, char* const argv[])
{
  yoxControl& ctrl = outMgr::ctrl;

  std::string yoxOutFileStrLbaseName;
  std::string yoxOutFileStrGbaseName;
  std::string yoxOutFileStrBaseName;

  opterr = 0; // disable getopt () generated error messages

  int ch;

  // a leading '+' is same as setting environment variable `POSIXLY_CORRECT'
  do
    {
      ch = getopt_long (argc, argv, "+:b:cdG:hjLlnp:StuV", opts, nullptr);
      if (ch == -1) break;

      switch (ch)
        {
        case PURE_PARSER_OPT: // --pure-parser, --reentrant-parser
          ctrl.allLfilesForReentrantParser = true;
          break;
        case HEADER_PREFIX_OPT: // --header-prefix basename
          yoxOutFileStrHbaseName = std::string (optarg);
          break;
        case HELP_OPT:
          printUsageText (std::cout);
          exit (EXIT_SUCCESS);
        case LEX_PREFIX_OPT: // --lex-prefix basename
          yoxOutFileStrLbaseName = std::string (optarg);
          break;
        case PARSE_PREFIX_OPT: // --parse-prefix basename
          yoxOutFileStrGbaseName = std::string (optarg);
          break;
        case FLEX_OPT: // --flex[=C|C++]
          ctrl.targetLexerGen = flexLexerGen;
          if (optarg == nullptr)
            {
              ctrl.targetLexerLanguage = languageC;
            }
          else if (strcasecmp (optarg, "c++") == 0)
            {
              ctrl.targetLexerLanguage = languageCXX;
            }
          else if (strcasecmp (optarg, "c") == 0)
            {
              ctrl.targetLexerLanguage = languageC;
            }
          else
            {
              const AGC::location loc; // dummy location
              yox_fatal (loc,
                         "unknown Flex target language "
                         "\"" + std::string (optarg) + "\"\n");
            }
#if (__APPLE__ && __clang__ && defined(__apple_build_version__) && \
     __clang_major__ > 11) ||                                      \
    (__GNUC__ > 6)
          // Apple Xcode 10 & 11 gcc, GNU 4, 5 & 6 gcc issue
          // a warning: declaration does not declare anything
          __attribute__ ((fallthrough));
#endif
        case REFLEX_OPT: // --reflex[=C++]
          ctrl.targetLexerGen = reflexLexerGen;
          if (optarg == nullptr)
            {
              ctrl.targetLexerLanguage = languageCXX;
            }
          else if (strcasecmp (optarg, "c++") == 0)
            {
              ctrl.targetLexerLanguage = languageCXX;
            }
          else
            {
              const AGC::location loc; // dummy location
              yox_fatal (loc,
                         "RE/flex does not support target language \"" +
                             std::string (optarg) + "\"\n");
            }
          break;
        case LEX_OPT: // --lex[=C|C++]
          ctrl.targetLexerGen = lexLexerGen;
          if (optarg == nullptr)
            {
              ctrl.targetLexerLanguage = languageC;
            }
          else if (strcasecmp (optarg, "c++") == 0)
            {
              ctrl.targetLexerLanguage = languageCXX;
            }
          else if (strcasecmp (optarg, "c") == 0)
            {
              ctrl.targetLexerLanguage = languageC;
            }
          else
            {
              const AGC::location loc; // dummy location
              yox_fatal (loc,
                         "unknown Lex target language "
                         "\"" + std::string (optarg) + "\"\n");
            }
          break;
        case BISON_OPT:     // --bison[=C|C++]
        case BISON_OLD_OPT: // --bison-2.3[=C|C++]
        case BTYACC_OPT:    // --btyacc[=C|C++]
        case BYACC_OPT:     // --byacc[=C|C++]
        case MSTA_OPT:      // --msta[=C|C++]
        case YACC_OPT:      // --yacc[=C|C++]
          switch (ch)
            {
#if BISON_TARGET_ONLY || ALL_TARGETS_STATIC || defined(USE_LT_DLOPEN)
            case BISON_OPT:
              ctrl.targetParserGen = bisonParserGen;
              break;
            case BISON_OLD_OPT:
              ctrl.targetParserGen = bisonParserGenOld;
              break;
#endif
#if BTYACC_TARGET_ONLY || ALL_TARGETS_STATIC || defined(USE_LT_DLOPEN)
            case BTYACC_OPT:
              ctrl.targetParserGen = btyaccParserGen;
              break;
#endif
#if BYACC_TARGET_ONLY || ALL_TARGETS_STATIC || defined(USE_LT_DLOPEN)
            case BYACC_OPT:
              ctrl.targetParserGen = byaccParserGen;
              break;
#endif
#if MSTA_TARGET_ONLY || ALL_TARGETS_STATIC || defined(USE_LT_DLOPEN)
            case MSTA_OPT:
              ctrl.targetParserGen = mstaParserGen;
              break;
#endif
#if YACC_TARGET_ONLY || ALL_TARGETS_STATIC || defined(USE_LT_DLOPEN)
            case YACC_OPT:
              ctrl.targetParserGen = yaccParserGen;
              break;
#endif
            default: // target parser generator not configured
              std::cerr << "Ox: \"" << argv[optind - 1]
                        << "\" not configured\n";
              printUsageText (std::cerr);
              exit (EXIT_FAILURE);
            }
          if (optarg == nullptr)
            {
              ctrl.targetParserLanguage = languageC;
            }
          else if (strcasecmp (optarg, "c++") == 0)
            {
              ctrl.targetParserLanguage = languageCXX;
            }
          else if (strcasecmp (optarg, "c") == 0)
            {
              ctrl.targetParserLanguage = languageC;
            }
          else
            {
              const AGC::location loc; // dummy location
              yox_fatal (loc,
                         "unknown target language "
                         "\"" + std::string (optarg) + "\"\n");
            }
          break;
        case LANGUAGE_OPT: // --language=[C|C++]
          if (strcasecmp (optarg, "c++") == 0)
            {
              ctrl.targetParserLanguage = languageCXX;
              ctrl.targetLexerLanguage = languageCXX;
            }
          else if (strcasecmp (optarg, "c") == 0)
            {
              ctrl.targetParserLanguage = languageC;
              ctrl.targetLexerLanguage = languageC;
            }
          else
            {
              const AGC::location loc; // dummy location
              yox_fatal (loc,
                         "unknown target language "
                         "\"" + std::string (optarg) + "\"\n");
            }
          break;

        case 'b': // --prefix basename
          yoxOutFileStrBaseName = std::string (optarg);
          break;
        case 'c':
          ctrl.suppressCycleReports = true;
          break;
        case 'd':
          ctrl.suppressPruning = true;
          break;
        case 'G':
          std::cerr << "Ox: command line lacks name for `Y-file\'\n";
          printUsageText (std::cerr);
          exit (EXIT_FAILURE);
        case 'h':
          ctrl.buildHeaderFile = true;
          break;
        case 'j':
          ctrl.saveBadOutputs = true;
          break;
        case 'L':
          ctrl.noLineDirectives = true;
          break;
        case 'l':
          ctrl.noLineDirectives = true;
          break;
        case 'n':
          ctrl.neutralizeSemActions = true;
          break;
        case 'p': // --name-prefix name
          ctrl.yoxSymbolPrefix = std::string (optarg);
          break;
        case 'S':
          ctrl.stripOxConstructs = true;
          break;
        case 't':
          ctrl.genTreeViz = true;
          break;
        case 'u':
          ctrl.showPTsizeStats = true;
          break;
        case 'V':
          printPatchlevel (std::cout);
          printVDATE (std::cout);
          exit (EXIT_SUCCESS);
        case ':': // missing option argument:
          switch (optopt)
            {
            case 'b':               // missing basename "prefix"
            case 'G':               // missing `L-file'
            case HEADER_PREFIX_OPT: // missing header "filename"
            case LEX_PREFIX_OPT:    // missing lexer "filename"
            case PARSE_PREFIX_OPT:  // missing parser "filename"
            default:                // something else is wrong
                ;
            }
#if (__APPLE__ && __clang__ && defined(__apple_build_version__) && \
     __clang_major__ > 11) ||                                      \
    (__GNUC__ > 6)
          // Apple Xcode 10 & 11 gcc, GNU 4, 5 & 6 gcc issue
          // a warning: declaration does not declare anything
          __attribute__ ((fallthrough));
#endif
        case '?': // unknown option
#if (__APPLE__ && __clang__ && defined(__apple_build_version__) && \
     __clang_major__ > 11) ||                                      \
    (__GNUC__ > 6)
          // Apple Xcode 10 & 11 gcc, GNU 4, 5 & 6 gcc issue an error:
          // a label can only be part of a statement and a declaration is
          // not a statement
          __attribute__ ((fallthrough));
#endif
        default:
          std::cerr << "Ox: error in command line option: " << argv[optind - 1]
                    << "\n";
          printUsageText (std::cerr);
          exit (EXIT_FAILURE);
        }
    } while (true);

  // don't generate `#line' directives if "-S"
  if (ctrl.stripOxConstructs)
    ctrl.noLineDirectives = true;

  if (!yoxOutFileStrBaseName.empty ())
    {
      if (yoxOutFileStrGbaseName.empty ()) // use <basename>
        yoxOutFileStrGbaseName = yoxOutFileStrBaseName;

      if (yoxOutFileStrLbaseName.empty ()) // use <basename>
        yoxOutFileStrLbaseName = yoxOutFileStrBaseName;

      if (yoxOutFileStrHbaseName.empty ()) // use <basename>
        yoxOutFileStrHbaseName = yoxOutFileStrBaseName;
    }

  // process remaining arguments

  argc -= optind;
  argv += optind;
  if (argc == 0)
    {
      std::cerr << "Ox: command line lacks name for `Y-file\'\n";
      printUsageText (std::cerr);
      exit (EXIT_FAILURE);
    }

  int i = 0;
  while ((i < argc) && (argv[i][0] == '-'))
    ++i;
  if (i == argc)
    {
      std::cerr << "Ox: command line lacks name for `Y-file\'\n";
      printUsageText (std::cerr);
      exit (EXIT_FAILURE);
    }

  std::string yoxInFileStrG = argv[i];

  std::string yoxOutFileStrG;
  if (!yoxOutFileStrGbaseName.empty ()) // use <parser-basename>
    yoxOutFileStrG = yoxOutFileStrGbaseName;
  else // use <default-basename>
    yoxOutFileStrG = yoxOutFileDefaultBaseName;

  std::string yoxOutFileStrH;
  if (!yoxOutFileStrHbaseName.empty ()) // use <header-basename>
    yoxOutFileStrH = yoxOutFileStrHbaseName;
  else // use <default-basename>
    yoxOutFileStrH = yoxOutFileDefaultBaseName;

  // set the parser file extension, based on the Y-file's extension
  addFilenameExt (yoxOutFileStrG, yoxInFileStrG, 'y');

  // set the gloval Y-file configuration parameters
  Yfile.inName = yoxInFileStrG;
  Yfile.outName = yoxOutFileStrG;

  // set the header file extension, based on the Y-file's extension
  addFilenameExt (yoxOutFileStrH, yoxInFileStrG, 'h');

  yoxOutH.open (yoxOutFileStrH);


  // process the L-file names

  while (++i < argc)
    {
      std::string arg = std::string (argv[i]);
      bool genScanner = false;

      if (arg == "-G")
        continue; // argument is exactly "-G"

      if (arg.substr (0, 2) == "-G") // argument is "-G<L-file>"
        {
          genScanner = true;
          arg = arg.substr (2); // drop the leading "-G"
        }
      else if (arg[0] == '-')
        {
          std::cerr
              << "Ox: command line option following non-option (ignored): "
              << arg << "\n";
          continue;
        }

      Lfile_specIO newLfile;

      newLfile.inName = arg;

      if (genScanner || std::string (argv[i - 1]) == "-G")
        newLfile.scanner = GEN_SPEC_LEXER;
      else
        newLfile.scanner = LEX_SPEC_LEXER;

      Lfile.push_back (newLfile);
    }

  if (Lfile.empty ())
    {
      std::cerr << "Ox: command line lacks name for `L-file\'\n";
      printUsageText (std::cerr);
      exit (EXIT_FAILURE);
    }

  int LfileSeq = 1;
  for (Lfile_specIO& currentLfile : Lfile)
    {
      std::string yoxOutFileStrL;

      if (!yoxOutFileStrLbaseName.empty ()) // use <lexer-basename>
        yoxOutFileStrL = yoxOutFileStrLbaseName;
      else // use <default-basename>
        yoxOutFileStrL = yoxOutFileDefaultBaseName;

      if (Lfile.size () > 1)
        yoxOutFileStrL.append (std::to_string (LfileSeq++));

      // set the lexer file extension, based on the L-file's extension
      addFilenameExt (yoxOutFileStrL, currentLfile.inName, 'l');

      currentLfile.outName = yoxOutFileStrL;
    }
}



void
openIOfiles (void)
{
  const yoxControl& ctrl = outMgr::ctrl;

  yoxOutGdefs.open ("@temp.yfiledefs.y");
  yoxOutGimport.open ("@temp.import.y");

  yoxOutHrequires.open ("@temp.requires.h");
  yoxOutHprovides.open ("@temp.provides.h");

  if (ctrl.buildHeaderFile)
    {
      if (!yoxOutFileStrHbaseName.empty ())
        {
          const std::string upper = make_upper (yoxOutFileStrHbaseName);

          yoxOutH.gen ()
              .put ("#ifndef __").put (upper).put ("_H\n")
              .put ("#define __").put (upper).put ("_H\n");
        }
      else
        {
          yoxOutH.gen ()
              .put ("#ifndef __OXOUT_H\n"
                    "#define __OXOUT_H\n");
        }

      yoxOutH.putPatchlevelComment ();

      yoxOutH.putParserHeaderText ();
    }

  if (ctrl.yoxSymbolPrefix != "yyy")
    {
      // generate user-specified prefixes for external symbols

      if (ctrl.genTreeViz)
        yoxOutHprovides.putTreeVizUserDefPrefix ();

      yoxOutHprovides.gen ()
          .put ("#define yyyabort " + yoxPrefix ("abort") + "\n")
          .put ("#define yyyinit " + yoxPrefix ("init") + "\n");
    }

  if (!ctrl.genTreeViz)
    yoxOutHprovides.gen ().put ("#define " + yoxPrefix ("init") + "()\n");
}



void
closeIOfiles (void)
{
  yoxOutG.close ();

  yoxOutL.close ();

  if (outMgr::ctrl.buildHeaderFile)
    yoxOutH.gen ().put ("#endif\n");

  yoxOutGdefs.close ();
  yoxOutGimport.close ();

  yoxOutH.close ();
  yoxOutHrequires.close ();
  yoxOutHprovides.close ();

  yoxOutGdefs.erase ();
  yoxOutGimport.erase ();

  if (!outMgr::ctrl.buildHeaderFile)
    yoxOutH.erase ();

  yoxOutHrequires.erase ();
  yoxOutHprovides.erase ();
}

void
eraseIOfiles (void)
{
  if (!outMgr::ctrl.removeOutput)
    return;

  yoxOutG.erase ();

  for (Lfile_specIO& l : Lfile)
    ::erase (l.outName);

  if (outMgr::ctrl.buildHeaderFile)
    yoxOutH.erase ();

  std::cerr << "Ox: did not generate output.\n";
}
