/* $Id: semtypes.hh 2834 2022-12-10 21:33:22Z teshields $ */

/* This file is part of Ox, an Attribute Grammar Compiling System
   that augments the specification functionality of Lex and Yacc, and
   input specification compatible lexer and parser generators.

   Copyright (C) 2019-2022  Thomas E. Shields

   This program is free software: you can redistribute it and/or modify
   it under the terms of the GNU General Public License as published by
   the Free Software Foundation, either version 3 of the License, or
   (at your option) any later version.

   This program is distributed in the hope that it will be useful,
   but WITHOUT ANY WARRANTY; without even the implied warranty of
   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
   GNU General Public License for more details.

   You should have received a copy of the GNU General Public License
   along with this program.  If not, see <http://www.gnu.org/licenses/>.  */

#ifndef SEMTYPES_HH_INCLUDED
#define SEMTYPES_HH_INCLUDED

#include <list>
#include <vector>

#include "location.hh"

// An attribute of a grammar symbol instance in the decorated parse tree is
// either inherited or synthesized, or else it is not yet known which
enum attrbDirection
{
  inherited,
  synthesized,
  undefined
};


// An attribute declared for a grammar symbol may have a copy rule annotation
// (@autoinh or @autosyn, @warn) that defines the automatic generation of copy
// rules for that attribute for that symbol.
struct attrbAnnotation
{
  enum attrbDirection direction;
  int lineno; // line# of definition, if defined, else -1

  bool autoWarn;

  attrbAnnotation ()
      : direction (undefined),
        lineno (-1),
        autoWarn (false)
  {}

  attrbAnnotation (enum attrbDirection d, int l, bool w)
      : direction (d),
        lineno (l),
        autoWarn (w)
  {}
};


// Each @attributes declaration defines a 'sort' (in the sense of 'kind')
// corresponding to a collection of attributes associated with each grammar
// symbol listed in the @attributes declaration.


// Each attribute in a 'sort' is represented by a `sortAttrb'.
struct sortAttrb
{
  std::string id; // attribute name

  attrbAnnotation
      autoDef; // copy rule annotation, if there is one for this attribute

  sortAttrb ()
      : id (""),
        autoDef (attrbAnnotation ())
  {}

  sortAttrb (const std::string& name, attrbAnnotation anno)
      : id (name),
        autoDef (anno)
  {}
};


// The collection of attributes in a 'sort' are represented by a (unique) vector
// of `sortAttrb'.
typedef std::vector<sortAttrb> sortAttrbSeq;


#define NOTYPE 0
#define firstSortTypeNum (NOTYPE + 1)

// The semantic representation of an attribute sort
struct sort
{
  int typeNum;          // the serial number of the sort
  std::string typeName; // the struct name for the sort
  int nAttrbs;
  sortAttrbSeq* attrbSeq;

  sort ()
      : // create undefined
        typeNum (NOTYPE),
        typeName (""),
        nAttrbs (0),
        attrbSeq (nullptr)
  {}
};


// The semantic representation of a collection of attribute sorts
typedef std::vector<sort> sortCollection;


// semantic representation of an attribute occurrence
struct attrbPair
{
  int pos; // symbol position of an attribute occurrence in rule
  int num; // serial number of an attribute in an occurrence
           // (= 'NIOsymAttrbNum' for an NIO pseudo attribute occurrence)

  inline attrbPair () = default;

  inline attrbPair (int p, int n)
      : pos (p),
        num (n)
  {}
};

// attribute occurrence with its location
struct attrbRefSeqElem
{
  attrbPair attrbRef;
  AGC::location loc;

  inline attrbRefSeqElem ()
      : attrbRef (attrbPair ()),
        loc (AGC::location ())
  {}

  inline attrbRefSeqElem (attrbPair a, AGC::location l)
      : attrbRef (a),
        loc (l)
  {}
};

// an attribute occurrence sequence in an attribute definition;
// used for processing the dependents & dependees in the definition
typedef std::list<attrbRefSeqElem> attrbRefSeq;


// a type sequence

enum kind_of
{
  is_ID,
  is_NAMESPACE_ID,
  is_TYPE_KEYWORD,
  is_literal
};

struct typeSeqElem
{
  std::string item;
  enum kind_of kind;

  inline typeSeqElem ()
      : item (std::string ()),
        kind (is_literal)
  {}

  inline typeSeqElem (std::string i)
      : item (i),
        kind (is_literal)
  {}

  inline typeSeqElem (std::string i, enum kind_of k)
      : item (i),
        kind (k)
  {}
};

typedef std::list<typeSeqElem> typeSeq;


// a token or symbol sequence
typedef std::list<std::string> symSeq;


// a {type sequence, symbol sequence} sequence
struct symDeclSeqElem
{
  typeSeq types;
  symSeq syms;

  inline symDeclSeqElem ()
      : types (typeSeq ()),
        syms (symSeq ())
  {}

  inline symDeclSeqElem (typeSeq t, symSeq s)
      : types (t),
        syms (s)
  {}
};

typedef std::list<symDeclSeqElem> symDeclSeq;


// a user argument sequence
struct userArgSeqElem
{
  typeSeq types;
  std::string userArg;

  inline userArgSeqElem ()
      : types (typeSeq ()),
        userArg (std::string ())
  {}

  inline userArgSeqElem (typeSeq t, std::string u)
      : types (t),
        userArg (u)
  {}
};

typedef std::list<userArgSeqElem> userArgSeq;

#endif // !SEMTYPES_HH_INCLUDED
