%{

/* RCS Log: scan.l,v
 * Revision 1.4  1989/03/27  16:29:06  kelvin
 * Fixed type conflict between yytext (unsigned char *) and cp (char *)
 *   in two places.
 *
 * Revision 1.3  89/03/26  18:30:39  kelvin
 * Changed Header to Log.
 * 
 */

/*
  Lexical Analyzer
*/
#include <stdio.h>

#include "sym.h"
#include "parser.h"

extern double atof();

static void comment(void);
static void count(void);
static int id(void);
static int dec_constant(void);
static int oct_constant(void);
static int hex_constant(void);
static void bad_char(void);

#if YYBTYACC && !YYBYACC
void yyerror(const char *msg, ...);
#else
void yyerror(const char *msg);
#endif

static int
  linenumber = 1,
  columnnumber = 0;

%}

letter			[a-zA-Z_]
digit			[0-9]
hex_digit		[a-fA-F0-9]
oct_digit		[0-7]
nonzero_digit		[1-9]
dec_constant		{nonzero_digit}{digit}*
hex_constant		(0x|0X){hex_digit}+
oct_constant		0{oct_digit}*

%%
[ \t]*			{ count(); /* ignore whitespace */ }
\n			{ linenumber++; columnnumber = 0; }
"/*"			{ comment(); }

int			{ count(); return TK_INT; }
extern			{ count(); return TK_EXTERN; }
struct			{ count(); return TK_STRUCT; }
if			{ count(); return TK_IF; }
else			{ count(); return TK_ELSE; }
for			{ count(); return TK_FOR; }
do			{ count(); return TK_DO; }
while			{ count(); return TK_WHILE; }
sizeof			{ count(); return TK_SIZEOF; }
break			{ count(); return TK_BREAK; }
return			{ count(); return TK_RETURN; }

({letter}|_)({letter}|{digit}|_)*		{ count(); return id(); }

{dec_constant}		{ count(); return dec_constant(); }
{hex_constant}		{ count(); return hex_constant(); }
{oct_constant}		{ count(); return oct_constant(); }


"++"			{ count(); return TK_INC; }
"--"			{ count(); return TK_DEC; }
"->"			{ count(); return TK_PTR; }
"||"			{ count(); return TK_OR; }
"<="			{ count(); return TK_LE; }
">="			{ count(); return TK_GE; }
"=="			{ count(); return TK_EQ; }
"!="			{ count(); return TK_NE; }
"&&"			{ count(); return TK_AND; }

"."			{ count(); return('.'); }
","			{ count(); return(','); }
";"			{ count(); return(';'); }
"{"			{ count(); return('{'); }
"}"			{ count(); return('}'); }
"("			{ count(); return('('); }
")"			{ count(); return(')'); }
"["			{ count(); return('['); }
"]"			{ count(); return(']'); }
"&"			{ count(); return('&'); }
"!"			{ count(); return('!'); }
"="			{ count(); return('='); }
"-"			{ count(); return('-'); }
"+"			{ count(); return('+'); }
"*"			{ count(); return('*'); }
"/"			{ count(); return('/'); }
"%"			{ count(); return('%'); }
"<"			{ count(); return('<'); }
">"			{ count(); return('>'); }

.			{ /* sop up any other characters */ bad_char(); }

%%

int yywrap(void)
{
  return(1);
}

/* read remainder of comments */
static void comment(void)
{
  char c;

  count();
  c = input();
  do {
    unput(c);
    while ((c = input()) != '*')
      if (!c) 
	yyerror("unterminated comment");
      else if (c == '\n') {
	linenumber++;
	columnnumber = 0;
      }
      else if (c == '\t')
	columnnumber += 8 - (columnnumber % 8);
      else
	columnnumber++;
    if (c)
      columnnumber++;		/* count the '*' */
  } while ((c = input()) && (c != '/'));
  if (!c)
    yyerror("unterminated comment");
  else
    columnnumber++;		/* count the closing '/' */
}

/* keep column numbers correct */
static void count(void)
{
  int i;

  for (i = 0; yytext[i] != '\0'; i++)
    if (yytext[i] == '\n')
      columnnumber = 0;
    else if (yytext[i] == '\t')
      columnnumber += 8 - (columnnumber % 8);
    else
      columnnumber++;
}

static int id(void)
{
  char *slookup();
  
  yylval.s = slookup(yytext, 1);
  return TK_IDENTIFIER;
}

/* int_constant - read an integer constant */
static int dec_constant(void)
{
  yylval.i = atoi(yytext);
  return(TK_ICONSTANT);
}

static int oct_constant(void)
{
  register char *cp;
  register int n;
  
  n = 0;
  for (cp = (char *) yytext + 1; *cp; )
    n = 8 * n + *cp++ - '0';
  yylval.i = n;
  return TK_ICONSTANT;
}

static int hex_constant(void)
{
  register char *cp;
  register int n;
  static char htod[] = {
    0, 1, 2, 3, 4, 5, 6, 7, 8, 9,-1,-1,-1,-1,-1,-1,
    -1,10,11,12,13,14,15,-1,-1,-1,-1,-1,-1,-1,-1,-1,
    -1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,
    -1,10,11,12,13,14,15,
  };	
  
  n = 0;
  for (cp = (char *) yytext + 2; *cp; )
    n = 16 * n + htod[*cp++ - '0'];
  yylval.i = n;
  return TK_ICONSTANT;
}

/* unrecognized character found */
static void bad_char(void)
{
  char buffer[80];
  
  sprintf(buffer, "unknown character 0%o (%c)", yytext[0], yytext[0]);
  yyerror(buffer);
}


#if YYBTYACC && !YYBYACC
void yyerror(const char *msg, ...)
#else
void yyerror(const char *msg)
#endif
{
  (void) fprintf(stderr, "line %d, col %d: %s\n",
		 linenumber, columnnumber, msg);
}
