#!/bin/bash
shopt -s extglob nullglob

# $Id: run-Ox-test.sh 2486 2022-04-27 01:05:30Z teshields $

# This file is part of Ox, an Attribute Grammar Compiling System
# that augments the specification functionality of Lex and Yacc, and
# input specification compatible lexer and parser generators.
#
# Copyright (C) 2017-2022  Thomas E. Shields
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.

DIFF_CMD="${DIFF}"
case "$build_os" in
    cygwin*)
	host_os="cygwin" ;;
    darwin*)
	host_os="macOS" ;;
    linux-gnu*)
	host_os="linux" ;;
    mingw*)
	DIFF_CMD="${DIFF} --strip-trailing-cr"
	host_os="mingw" ;;
    msys*)
	host_os="msys" ;;
    solaris*)
	host_os="sunOS" ;;
    *)
	host_os="unknown" ;;
esac

${OX} -V | ${DIFF_CMD} - ${srcdir}/ver.bak > ver.diff
if test -s ver.diff
then echo "Cannot run Ox test $1: "
     echo "  Testsuite version does not match Ox version"
     cat ver.diff
     echo "  for"
     echo "  Ox = ${OX}"
     rm -f ver.diff
     exit 99 # hard error!
fi
rm -f ver.diff

header_pattern=`grep Ox ${srcdir}/ver.bak`

subdir=${1##*/}
if [ -d ${srcdir}/$subdir ]
then

     # clean out everything first
     ${MAKE} --directory=$subdir ${AM_MAKEFLAGS} clean > /dev/null 2>&1

     # run the `check' target; quit if there are Ox errors (should be none), or
     # if there are errors building the `gc' executable (if build rules are present)
     if ${MAKE} --directory=$subdir ${AM_MAKEFLAGS} check check-options > $subdir/makerr 2>&1
     then if [ -s $subdir/makerr ]
	  then echo "Messages making in $PWD/$subdir:"
	       cat $subdir/makerr
	       echo ""
	  fi
	  rm -f $subdir/makerr
     else echo "Error(s) making in $PWD/$subdir:"
	  echo "Ox error(s):"
	  cat $subdir/makerr
	  echo ""
	  rm -f $subdir/makerr
	  exit 1
     fi

     failed=0

     if [ "$subdir" = "syntax-errors" -o "$subdir" = "semantic-errors" ]
     then # check for changes in the Ox stderr files (should be none)
	  files=`echo $subdir/*.errout`
	  if [ -n "$files" ]
	  then for j in $files ; do
		  if [ -f ${srcdir}/$j.bak ]
		  then # `sed' alternatives below handle VPATH builds; no-op otherwise
		       if [ "$host_os" == "mingw" ]
		       then
			    mingw_prefix=`cygpath -m /`
			    mingw_prefix=`dirname $mingw_prefix`/`basename $mingw_prefix`
			    sed -e "s=${mingw_prefix}==g" -e "s=${srcdir}/$subdir/==g" -e "s=\.\./==g" $j > $j.cmp
		       else
			    sed -e "s=${srcdir}/$subdir/==g" -e "s=\.\./==g" $j > $j.cmp
		       fi
		       # compare the baseline file with the Ox stderr file
		       echo -n "Checking $PWD/$j: "
		       ${DIFF_CMD} ${srcdir}/$j.bak $j.cmp > $j.diff
		       [ $? -eq 2 ] && exit 2
		       if [ -s $j.diff ]
		       then echo "changes:"
			    cat $j.diff
			    echo ""
			    failed=1
		       else echo "no change"
		       fi
		       rm -f $j.cmp $j.diff
		  else echo "Missing master copy ${srcdir}/$j.bak"
		       failed=1
		  fi
	       done
	  else echo "Missing Ox output files in $PWD/$subdir"
	       failed=1
	  fi
     else # check for changes in the code generated by Ox (should be none)
      files=`echo $subdir/?(strip.|neuter.)lexer?([0-9]).?(line?([0-9]).)l?(l|pp) $subdir/?(strip.|neuter.)parser.?(line.)y?(y|pp) $subdir/semtypes.?(line.)h?(h|pp)`
	  if [ -n "$files" ]
	  then for j in $files ; do
		  if [ -f ${srcdir}/$j.bak ]
		  then # next line handles VPATH builds; no-op otherwise
		       sed "/^ *#line/s=\".*/$subdir/=\"=" $j > $j.cmp
		       # next 2 lines ignore changes in SVN `Id' keyword lines
		       grep -v '.*\$Id:' $j.cmp > $j.tmp
		       grep -v '.*\$Id:' ${srcdir}/$j.bak > $j.bak.tmp
		       # compare the baseline file with the file Ox generated
		       echo -n "Checking $PWD/$j: "
		       ${DIFF_CMD} $j.bak.tmp $j.tmp > $j.diff
		       [ $? -eq 2 ] && exit 2
		       if [ -s $j.diff ]
		       then echo "changes:"
			    cat $j.diff
			    echo ""
			    failed=1
		       else echo "no change"
		       fi
		       rm -f $j.cmp $j.tmp $j.bak.tmp $j.diff
		  else echo "Missing master copy ${srcdir}/$j.bak"
		       failed=1
		  fi
	       done
	  else echo "Missing Ox output files in $PWD/$subdir"
	       failed=1
	  fi
     fi

     if [ -x $subdir/gc${EXEEXT} ]
     then # run the `gc' executable on any test cases present
	  if [ -d ${srcdir}/$subdir/gct ]
	  then mkdir -p $subdir/gct
	       for j in `echo ${srcdir}/$subdir/gct/t+([0-9])` ; do
		  k=${j##*/}
		  $subdir/gc${EXEEXT} < $j > $subdir/gct/$k.o 2>&1
		  if [ "$subdir" = "gppl" ]
		  then # these test cases only generate error messages;
		       # we don't need the generated GPPL code (if any)
		       rm -f g.c # generated code is in cwd
		  fi
		  if [ -f $j.ob ]
		  then # compare the expected output with the file just written
		       echo -n "Checking $PWD/$subdir/gct/$k.o: "
		       ${DIFF_CMD} $j.ob $subdir/gct/$k.o > $subdir/gct/$k.diff
		       [ $? -eq 2 ] && exit 2
		       if [ -s $subdir/gct/$k.diff ]
		       then echo "changes:"
			    cat $subdir/gct/$k.diff
			    echo ""
			    failed=1
		       else echo "no change"
		       fi
		       rm -f $subdir/gct/$k.diff
		  else echo "Missing master copy $j.ob"
		       failed=1
		  fi
	       done
	  fi
	  if [ -d ${srcdir}/$subdir/gct2 ]
	  then # for these test cases, we want to test the generated GPPL code
	       for i in $subdir/gct2/* ; do
		  [ -d $i ] || continue

		  # check for changes in the GPPL generated code (should be none)
		  if [ -f ${srcdir}/$i/g.c.bak ]
		  then # compare the baseline file with the file just generated
		       echo -n "Checking $PWD/$i/g.c: "
		       ${DIFF_CMD} ${srcdir}/$i/g.c.bak $i/g.c > $i/g.c.diff
		       [ $? -eq 2 ] && exit 2
		       if [ -s $i/g.c.diff ]
		       then echo "changes:"
			    cat $i/g.c.diff
			    echo ""
			    failed=1
		       else echo "no change"
		       fi
		       rm -f $i/g.c.diff
		  else echo "Missing master copy ${srcdir}/$i/g.c.bak"
		       failed=1
		  fi

		  # process the subsidiary test cases
		  if [ -d ${srcdir}/$i/gct ]
		  then mkdir -p $i/gct
		       for j in `echo ${srcdir}/$i/gct/t+([0-9])` ; do
			  k=${j##*/}
			  $i/gc${EXEEXT} < $j > $i/gct/$k.o 2>&1
			  [ $? -ne 0 ] && exit 1
			  if [ -f $i/gct/$k.o ]
			  then if [ -f $j.ob ]
			       then # compare the expected output with the file just written
				    echo -n "Checking $PWD/$i/gct/$k.o: "
				    ${DIFF_CMD} $j.ob $i/gct/$k.o > $i/gct/$k.diff
				    res=$?
				    if [ $res -ne 0 ]
				    then
					 if [ -f $j.$host_os.ob ]
					 then
					      ${DIFF_CMD} $j.$host_os.ob $i/gct/$k.o > $i/gct/$k.diff
					 else
					      [ $res -eq 2 ] && exit 2
					 fi
				    fi
				    [ $? -eq 2 ] && exit 2
				    if [ -s $i/gct/$k.diff ]
				    then echo "changes:"
					 cat $i/gct/$k.diff
					 echo ""
					 failed=1
				    else echo "no change"
				    fi
				    rm -f $i/gct/$k.diff
			       else echo "Missing master copy $j.ob"
				    failed=1
			       fi
			  else echo "Expected output file $i/gct/$k.o missing"
			       failed=1
			  fi
		       done
		  fi
	       done
	  fi
     fi

     if [ -x $subdir/gc_treeviz${EXEEXT} ]
     then # run the `gc_treeviz' executable on any test cases present
	  if [ -d ${srcdir}/$subdir/gct ]
	  then mkdir -p $subdir/gct
	       for j in `echo ${srcdir}/$subdir/gct/t+([0-9])` ; do
		  k=${j##*/}
		  $subdir/gc_treeviz${EXEEXT} -o $subdir/gct/$k.gv $j > $subdir/gct/$k.o 2>&1
		  if [ "$subdir" = "gppl" ]
		  then # these test cases only generate error messages;
		       # we don't need the generated GPPL code (if any)
		       rm -f g.c # generated code is in cwd
		  fi
		  if [ -f $subdir/gct/$k.o ]
		  then if [ -f $j.ob ]
		       then # compare the expected output with the file just written
			    echo -n "Checking $PWD/$subdir/gct/$k.o: "
			    ${DIFF_CMD} $j.ob $subdir/gct/$k.o > $subdir/gct/$k.diff
			    [ $? -eq 2 ] && exit 2
			    if [ -s $subdir/gct/$k.diff ]
			    then echo "changes:"
				 cat $subdir/gct/$k.diff
				 echo ""
				 failed=1
			    else echo "no change"
			    fi
			    rm -f $subdir/gct/$k.diff
		       else echo "Missing master copy $j.ob"
			    failed=1
		       fi
		  else echo "Expected output file $subdir/gct/$k.o missing"
		       failed=1
		  fi
		  # check the DOT-language output
		  if [ -f $j.gv ]
		  then if [ -f $j.gv.ob ]
		       then # compare the expected output with the file just written
			    echo -n "Checking $PWD/$subdir/gct/$k.gv: "
			    # ignore all whitespace (portability)
			    ${DIFF_CMD} -w $j.gv.ob $subdir/gct/$k.gv > $subdir/gct/$k.gv.diff
			    [ $? -eq 2 ] && exit 2
			    if [ -s $subdir/gct/$k.gv.diff ]
			    then echo "changes:"
				 cat $subdir/gct/$k.gv.diff
				 echo ""
				 failed=1
			    else echo "no change"
			    fi
			    rm -f $subdir/gct/$k.gv.diff
		       else echo "Missing master copy $j.gv.ob"
			    failed=1
		       fi
		  fi
	       done
	  fi
     fi
     exit $failed
else
     echo "Error invoking $0"
     exit 99 # hard error!
fi
